// Copyright 1992, 1997-2000 by Jon Dart.  All Rights Reserved.

#ifndef _BEARING_H
#define _BEARING_H
#include "board.h"
#include "move.h"

extern const int Direction[2];

extern const byte KnightSquares[64][8];

extern const byte KingSquares[64][8];

extern const byte BishopSquares[64][32];

extern const byte RookSquares[64][32];

extern const int RankIncr;	// add this to move 1 rank

#define DIAG_ATTACKS_LEN 220

class Bearing
{
     // Finds pseudo-legal moves for individual pieces, plus
     // related functions.

     public:
     
     enum { MaxBearSq = 28 };	     
     
     static Bitmap knight_attacks[64];
     static Bitmap pawn_attacks[64][2];
     static Bitmap king_attacks[64];
     static Bitmap unrot90[256][8];
     
     // support for "compact attacks"
     static byte whichAttack[256][8];
     static Bitmap rankAttacks[64];
     static Bitmap rankAttacksUp[64];
     static Bitmap rankAttacksDown[64];
     static byte fileAttacks[64];
     static byte fileAttacksUp[64];
     static byte fileAttacksDown[64];
     static Bitmap diagA1Attacks[DIAG_ATTACKS_LEN];
     static Bitmap diagA8Attacks[DIAG_ATTACKS_LEN];

     static int mobility[64];
     static int diagonal_length[64];
     static int Directions[64][64];
     static Bitmap file_mask[8];
     static Bitmap rank_mask[8];
     static Bitmap diag_a1_mask[64];
     static Bitmap diag_a8_mask[64];
     static Bitmap white_squares;
     static Bitmap black_squares;
     // squares that could attack a given square on rank or file
     static Bitmap rank_file_attack[64];
     // squares that could attack a given square on the diagonal
     static Bitmap diag_attack[64];

     // mask for e.p. checking
     static Bitmap ep_mask[8][2];
     
     // Initialize the bitmaps.  Call before using this class.
     static void init();

     static unsigned BearSq( const Board &board,
	 const Square loc, Square *squares);
     // Returns all squares to which a piece at location 'loc' can move.
     // It does not include castling moves, but does include en passant
     // captures and pawn promotions.  It doesn't check for pins, 
     // and hence may return some illegal moves.
	     
     static unsigned Attack( const Board &board,
	 const Square loc, const ColorType side,
	 Square *squares);
     // returns a list of all squares on which pieces of color "side"
     // reside which attack the piece at "loc".  The function return
     // value is the number of such pieces.  If "loc" is empty,
     // returns all squares from which pieces can move to "loc".
     // En passant captures are not included.

     static int Pinned( const Board &board, const Square loc,
	     Piece &PinnedByPiece, 
	     Square &PinnedBySquare,
	     int &dir);
     // returns TRUE if the piece at location 'square' is pinned. If it
     // is pinned, also returns PinnedByPiece = piece pinning piece at
     // 'square', PinnedBySquare = pinning piece's square, dir =
     // direction of pin.	     
     // We could call Attacks to get this information, but this is faster.
	     
};

inline Bitmap FileAttacks(const Board &board,Square sq)
{
      int f = File(sq);
      byte atck = Bearing::whichAttack[board.occupiedR90.rank_bits(8-f)][((int)sq)/8];
      return Bitmap(Bearing::rankAttacks[atck]).shl_all(8-f);
}

inline Bitmap FileAttacksDown(const Board &board,Square sq)
{
      int f = File(sq);
      byte atck = Bearing::whichAttack[board.occupiedR90.rank_bits(8-f)][((int)sq)/8];
      return Bitmap(Bearing::rankAttacksDown[atck]).shl_all(8-f);
}

inline Bitmap FileAttacksUp(const Board &board,Square sq)
{
      int f = File(sq);
      byte atck = Bearing::whichAttack[board.occupiedR90.rank_bits(8-f)][((int)sq)/8];
      return Bitmap(Bearing::rankAttacksUp[atck]).shl_all(8-f);
}

inline Bitmap RankAttacks(const Board &board,Square sq)
{
      Bitmap value;
      ((char*)&value)[sq/8] = 
        Bearing::fileAttacks[
        Bearing::whichAttack[board.all_occupied.rank_bits((int)sq/8)][File(sq)-1]];
      return value;
}

inline Bitmap RankAttacksRight(const Board &board,Square sq)
{
      Bitmap value;
      ((char*)&value)[sq/8] = Bearing::fileAttacksDown[
        Bearing::whichAttack[board.all_occupied.rank_bits((int)sq/8)][File(sq)-1]];
      return value;
}

inline Bitmap RankAttacksLeft(const Board &board,Square sq)
{
      Bitmap value;
      ((char*)&value)[sq/8] = 
        Bearing::fileAttacksUp[
        Bearing::whichAttack[board.all_occupied.rank_bits((int)sq/8)][File(sq)-1]];
      return value;
}

inline Bitmap DiagAtcksA1(const Board &board,Square sq)
{
   const register DiagInfo2 *info2 = &DiagInfoA1[sq];
   return info2->attacks[
     Bearing::whichAttack[(int)((((byte*)(&board.occupiedDa1))[info2->byte_shift])
       >> info2->bit_shift) & info2->mask][info2->bit]];
}

inline Bitmap DiagAtcksA8(const Board &board,Square sq)
{
   const register DiagInfo2 *info2 = &DiagInfoA8[sq];
   return info2->attacks[
     Bearing::whichAttack[(int)((((byte*)(&board.occupiedDa8))[info2->byte_shift])
        >> info2->bit_shift) & info2->mask][info2->bit]];
}

inline Bitmap DiagAtcksA1Upper(const Board &board,Square sq)
{
   const register DiagInfo2 *info2 = &DiagInfoA1[sq];
   return Bitmap(info2->attacks[
     Bearing::whichAttack[(int)((((byte*)(&board.occupiedDa1))[info2->byte_shift])
       >> info2->bit_shift) & info2->mask][info2->bit]].data & info2->upperMask.data);
}

inline Bitmap DiagAtcksA1Lower(const Board &board,Square sq)
{
   const register DiagInfo2 *info2 = &DiagInfoA1[sq];
   return Bitmap(info2->attacks[
     Bearing::whichAttack[(int)((((byte*)(&board.occupiedDa1))[info2->byte_shift])
       >> info2->bit_shift) & info2->mask][info2->bit]].data & info2->lowerMask.data);
}

inline Bitmap DiagAtcksA8Upper(const Board &board,Square sq)
{
   const register DiagInfo2 *info2 = &DiagInfoA8[sq];
   return Bitmap(info2->attacks[
     Bearing::whichAttack[(int)((((byte*)(&board.occupiedDa8))[info2->byte_shift])
        >> info2->bit_shift) & info2->mask][info2->bit]].data & info2->upperMask.data);
}

inline Bitmap DiagAtcksA8Lower(const Board &board,Square sq)
{
   const register DiagInfo2 *info2 = &DiagInfoA8[sq];
   return Bitmap(info2->attacks[
     Bearing::whichAttack[(int)((((byte*)(&board.occupiedDa8))[info2->byte_shift])
        >> info2->bit_shift) & info2->mask][info2->bit]].data & info2->lowerMask.data);
}

#endif
