// Copyright 1994 by Jon Dart. All Rights Reserved.
// $Revision: 1.4 $
//
#include "movegen.h"
#include "bearing.h"
#include "attacke.h"
#include "util.h"
#include "debug.h"
#include "moveord.h"
#include "config.h"
#include "bearing.h"
#include <iostream>
#include <fstream>

extern const int Direction[2];

Move_Generator::Move_Generator(const Board & ABoard,
                               const unsigned curr_ply, const Move pvMove)
:board(ABoard), my_ply(curr_ply),
moves_generated(0)
{
}

Move_Generator::~Move_Generator()
{
}

static inline void swap( Move moves[], const int i, const int j)
{
   Move tmp = moves[j];
   moves[j] = moves[i]; 
   moves[i] = tmp;
}

int Move_Generator::
Generate_Moves(Move * moves, int repeatable)
{
   unsigned NumMoves  = 0;
   if (board.CheckStatus() == Board::InCheck)
   {
      NumMoves = Check_Evasions(moves,0);
   }
   else
   {
      NumMoves = 0;
      // castling moves
      const ColorType side = board.Side();
      Board::CastleType CS = board.CastleStatus(side);
      if ((CS == Board::CanCastleEitherSide) ||
          (CS == Board::CanCastleKSide))
      {
         const Square kp = board.KingPos(side);
#ifdef _DEBUG
         if (side == White)
         {
            ASSERT(kp == 60);
            ASSERT(board[kp+3] == whiteRook);
         }
         else
         {
            ASSERT(kp == 4);
            ASSERT(board[kp+3] == blackRook);
         }
#endif
         if (board[kp + 1] == EmptyPiece &&
             board[kp + 2] == EmptyPiece &&
             !board.any_attacks(kp,OppositeColor(side)) &&
             !board.any_attacks(kp + 1,OppositeColor(side)) &&
             !board.any_attacks(kp + 2,OppositeColor(side)))
            // can castle
            moves[NumMoves++] = CreateMove(kp, kp+2, InvalidPiece,
                                           King,EmptyPiece, KCastle);
      }
      if ((CS == Board::CanCastleEitherSide) ||
          (CS == Board::CanCastleQSide))
      {
         const Square kp = board.KingPos(side);
         if (board[kp - 1] == EmptyPiece &&
             board[kp - 2] == EmptyPiece &&
             board[kp - 3] == EmptyPiece && 
             !board.any_attacks(kp,OppositeColor(side)) &&
             !board.any_attacks(kp - 1,OppositeColor(side)) &&
             !board.any_attacks(kp - 2,OppositeColor(side)))
            // can castle
            moves[NumMoves++] = CreateMove(kp, kp-2, InvalidPiece,
                                           King, EmptyPiece, QCastle);
      }
      // other moves
      static Square squares[Bearing::MaxBearSq];
      Bitmap pieces(board.occupied[board.Side()]);
      for (;;)
      {
         Square loc = pieces.first_one(); //pieces.Next();
         if (loc == InvalidSquare)
            break;
         int n = Bearing::BearSq(board, loc, squares);
         const PieceType piecemoved = TypeOfPiece(board[loc]);
         for (int j = 0; j < n; j++)
         {
            const Square dest = squares[j];
            const PieceType capture = TypeOfPiece(board[dest]);
            if (piecemoved == Pawn)
            {
               if (Rank(dest,board.Side())==8)
               {
                  moves[NumMoves++] = CreateMove(loc, dest, Queen,
                                                 Pawn, capture, Promotion);
                  moves[NumMoves++] = CreateMove(loc, dest, Rook,
                                                 Pawn, capture, Promotion);
                  moves[NumMoves++] = CreateMove(loc, dest, Bishop,
                                                 Pawn, capture, Promotion);
                  moves[NumMoves++] = CreateMove(loc, dest, Knight,
                                                 Pawn, capture, Promotion);
               }
               else if (capture == EmptyPiece && File(dest) != File(loc))
                  moves[NumMoves++] = CreateMove(loc, dest, InvalidPiece,
                                                 piecemoved, Pawn, EnPassant);
               else
                  moves[NumMoves++] = CreateMove(loc, dest, InvalidPiece,
                                                 piecemoved, capture, Normal);
            }
            else
               moves[NumMoves++] = CreateMove(loc, dest, InvalidPiece,
                                              piecemoved, capture, Normal);
         }
         pieces.clear(loc);
      }
   }      
   if (repeatable)
   {
      static int scores[MaxMoves];
      for (unsigned i = 0; i < NumMoves; i++)
      {
         scores[i] = (int)StartSquare(moves[i]) + (int)(DestSquare(moves[i]) << 7);
         switch (PromoteTo(moves[i]))
         {
         case Queen:
            scores[i] &= 0xB000;
            break;
         case Rook:
            scores[i] &= 0x8000;
            break;
         case Bishop:
            scores[i] &= 0x4000;
            break;
         default:
            break;
         }
      }
      Move_Ordering::sort_moves(moves, scores, NumMoves);
   }
   return NumMoves;
}

#define RBAttacks(board, loc, data, moves, NumMoves) \
    Square i; \
    while ((i = *data) != 255 && board[i] == EmptyPiece) \
    { \
       ++data; \
    } \
    if (i != 255 && PieceColor(board[i]) != side) \
       moves[NumMoves++] = \
          CreateMove(loc,i,InvalidPiece,TypeOfPiece(board[loc]), \
          TypeOfPiece(board[i]),Normal) \

static void
PawnAttacks(const Board & board,
            const ColorType side, const Square i,
            Move *moves, unsigned &NumMoves, int ply)
{
   Square dest;
   Piece piece;

   // check for possible captures
   if (File(i) != 1)
   {
      dest = (side == White) ? i - (RankIncr + 1) :
             i + (RankIncr - 1);
      piece = board[dest];
      if (piece != EmptyPiece && (PieceColor(piece) != board.Side()))
      {
         if (Rank(i,side) == 7)
         {
            moves[NumMoves++] =
            CreateMove(i,dest,Queen,Pawn,TypeOfPiece(piece),Promotion);
            moves[NumMoves++] =
            CreateMove(i,dest,Knight,Pawn,TypeOfPiece(piece),Promotion);
            if (ply == 0)
            {
               moves[NumMoves++] =
               CreateMove(i,dest,Rook,Pawn,TypeOfPiece(piece),Promotion);
               moves[NumMoves++] =
               CreateMove(i,dest,Bishop,Pawn,TypeOfPiece(piece),Promotion);
            }
         }
         else
            moves[NumMoves++] =
            CreateMove(i,dest,InvalidPiece,Pawn,TypeOfPiece(piece),Normal);
      }
   }
   if (File(i) != 8)
   {
      dest = (side == White) ? i - (RankIncr - 1) :
             i + (RankIncr + 1);
      piece = board[dest];
      if (piece != EmptyPiece && (PieceColor(piece) != board.Side()))
      {
         if (Rank(i,side) == 7)
         {
            moves[NumMoves++] =
            CreateMove(i,dest,Queen,Pawn,TypeOfPiece(piece),Promotion);
            moves[NumMoves++] =
            CreateMove(i,dest,Knight,Pawn,TypeOfPiece(piece),Promotion);
            if (ply == 0)
            {
               moves[NumMoves++] =
               CreateMove(i,dest,Rook,Pawn,TypeOfPiece(piece),Promotion);
               moves[NumMoves++] =
               CreateMove(i,dest,Bishop,Pawn,TypeOfPiece(piece),Promotion);
            }
         }
         else
            moves[NumMoves++] =
            CreateMove(i,dest,InvalidPiece,Pawn,TypeOfPiece(piece),Normal);
      }
   }
   if (Rank(i,side) == 7)
   {
      dest = (side == White) ? i - RankIncr : i + RankIncr;
      piece = board[dest];
      if (piece == EmptyPiece)
      {
         // promotion
         moves[NumMoves++] =
         CreateMove(i,dest,Queen,Pawn,EmptyPiece,Promotion);
         moves[NumMoves++] =
         CreateMove(i,dest,Knight,Pawn,EmptyPiece,Promotion);
         if (ply == 0)
         {
            moves[NumMoves++] =
            CreateMove(i,dest,Rook,Pawn,TypeOfPiece(piece),Promotion);
            moves[NumMoves++] =
            CreateMove(i,dest,Bishop,Pawn,TypeOfPiece(piece),Promotion);
         }
      }
   }
   Square epsq = board.EnPassantSq();
   if (!IsInvalid(epsq))
   {
      ASSERT(TypeOfPiece(board[epsq])==Pawn);
      if ((File(i) != 8 && i + 1 == epsq) || 
          (File(i) != 1 && i - 1 == epsq))
      {
         Square dest = epsq + RankIncr * Direction[side];
         if (board[dest] == EmptyPiece)
            moves[NumMoves++] =
            CreateMove(i,dest,InvalidPiece,Pawn,Pawn,
                       EnPassant);
      }
   }
}

int Move_Generator::Generate_Captures(Move * moves)
{
   unsigned NumMoves;
   NumMoves = 0;

   const ColorType side = board.Side();
   Bitmap pieces(board.occupied[side]);
   for (;;)
   {
      Square loc = pieces.first_one();
      if (loc == InvalidSquare)
         break;
       int j;
       const ColorType side = PieceColor(board[loc]);
      switch (TypeOfPiece(board[loc]))
      {
      case EmptyPiece:
         break;
      case Pawn:
         PawnAttacks(board, side,
                     loc, moves, NumMoves, my_ply);
         break;
      case Knight:
         {
            Bitmap dests(Bearing::knight_attacks[loc]);
            if (side == White)
               dests.And(board.occupied[Black]);
            else
               dests.And(board.occupied[White]);
            Bitmap it(dests);
            Square sq;
            for (;(sq = it.first_one()) != InvalidSquare;it.clear(sq))
            {
               moves[NumMoves++] = 
               CreateMove(loc,sq,InvalidPiece,Knight,
                          TypeOfPiece(board[sq]),Normal);
            }
         }
         break;
      case Bishop:
         for (j = 0; j < 4; j++)
         {
            const byte *data = BishopSquares[loc] + (j*8);
            RBAttacks(board, loc, data, moves, NumMoves);
         }
         break;
      case Rook:
         for (j = 0; j < 4; j++)
         {
            const byte *data = RookSquares[loc] + (j*8);
            RBAttacks(board, loc, data, moves, NumMoves);
         }
         break;
      case Queen:
         for (j = 0; j < 4; j++)
         {
            const byte *data = BishopSquares[loc] + (j*8);
            RBAttacks(board, loc, data, moves, NumMoves);
         }
         for (j = 0; j < 4; j++)
         {
            const byte *data = RookSquares[loc] + (j*8);
            RBAttacks(board, loc, data, moves, NumMoves);
         }
         break;
      case King:
         {
            Bitmap dests(Bearing::king_attacks[loc]);
            if (side == White)
               dests.And(board.occupied[Black]);
            else
               dests.And(board.occupied[White]);
            dests.andNot(Bearing::king_attacks[board.KingPos(
             board.OppositeSide())]);
            Bitmap it(dests);
            Square sq;
            for (;(sq = it.first_one()) != InvalidSquare;it.clear(sq))
            {
               moves[NumMoves++] = 
               CreateMove(loc,sq,InvalidPiece,King,
                          TypeOfPiece(board[sq]),Normal);
            }
         }
         break;
      case InvalidPiece:
         break;
      }
      pieces.clear(loc);
   }
   moves_generated += NumMoves;
   return NumMoves;
}

static int is_pinned(const Board & board, const Square source,
                      const Square dest)
{
   int pin = 0;
   if (board.any_attacks(source, board.OppositeSide()))
   {
      // It is possible that the piece we plan to move is pinned.
      Piece PinnedByPiece;
      Square PinnedBySquare;
      int dir;
      if (Bearing::Pinned(board, source, PinnedByPiece,
                          PinnedBySquare, dir))
      {
         // This code checks for moves in the direction of the pin,
         // which are ok:
         int dir1 = Util::Abs((int) source - (int) dest);
         int dir2 = Util::Abs(dir);
         if (dir2 == 1 && Rank(source,White) != Rank(dest,White))
	   pin = 1;
         else if (dir2 == RankIncr && File(source) != File(dest))
	   pin = 1;
         else if (dir1 % dir2)
	   pin = 1;
      }
   }
   return pin;
}

unsigned Move_Generator::Check_Evasions(Move * moves, int captures_only)
{
   int num_moves = 0;
   unsigned n;
   unsigned i;
   static Square squares[36];
   Square source;
   const Square kp = board.KingPos(board.Side());
   Bitmap king_attacks(board.calc_attacks(kp, board.OppositeSide()));
   ASSERT(!king_attacks.is_clear());
   int num_attacks = 0;
   int fo = king_attacks.first_one();
   num_attacks = king_attacks.bit_count();
   if (num_attacks == 1)
   {
      // try to capture checking piece
      source = (Square)fo;

      ASSERT(source != InvalidSquare);
      Bitmap atcks(board.calc_attacks(source,board.Side()));
      Square sq;
      for (;(sq = atcks.first_one()) != InvalidSquare;atcks.clear(sq))
      {
         PieceType capturingPiece = TypeOfPiece(board[sq]);
         if (capturingPiece == King)
         {
            // We can capture with the king only if the piece
            // checking us is undefended.  But always allow a
            // capture *of* the king - for illegal move detection.
            if (TypeOfPiece(board[source]) == King ||
                !board.any_attacks(source, board.OppositeSide()))
            {
               moves[num_moves++] = CreateMove(board, sq, source, InvalidPiece);
            }
         }
         else
         {
            if (!is_pinned(board, sq, source))
            {
               if (capturingPiece == Pawn &&
                   Rank(source,board.Side()) == 8)
               {
                  moves[num_moves++] = CreateMove(board,
                                                  sq, source, Queen);
                  moves[num_moves++] = CreateMove(board,
                                                  sq, source, Rook);
                  moves[num_moves++] = CreateMove(board,
                                                  sq, source, Knight);
                  moves[num_moves++] = CreateMove(board,
                                                  sq, source, Bishop);
               }
               else
                  moves[num_moves++] = CreateMove(board,sq, source, InvalidPiece);
            }
         }
      }
      // Bearing::Attack does not return en passant captures, so try
      // this as a special case
      if (board.EnPassantSq() == source)
      {
         Square dest = source + RankIncr * Direction[board.Side()];
         Piece myPawn = MakePiece(Pawn,board.Side());
         if (File(source) != 8 && board[source + 1] == myPawn)
         {
            Piece tmp = board[source];
            Piece &place = (Piece &)board[source];
            place = EmptyPiece; // imagine me gone
            if (!is_pinned(board, source + 1, dest))
               moves[num_moves++] = CreateMove(source + 1, dest, InvalidPiece, Pawn,
                                               Pawn, EnPassant);
            place = tmp;
         }
         if (File(source) != 1 && board[source - 1] == myPawn)
         {
            Piece tmp = board[source];
            Piece &place = (Piece &)board[source];
            place = EmptyPiece; // imagine me gone
            if (!is_pinned(board, source - 1, dest))
               moves[num_moves++] = CreateMove(source - 1, dest, InvalidPiece, Pawn,
                                               Pawn, EnPassant);
            place = tmp;
         }
      }
      if (captures_only)
         return num_moves;
      // try to interpose a piece
      if (TypeOfPiece(board[source]) != Knight &&
          TypeOfPiece(board[source]) != Pawn)
      {
         static Square btwn_squares[8];
         unsigned nbsq = board.Between(source, kp, btwn_squares);
         ASSERT(nbsq<8);
         for (i = 0; i < nbsq; i++)
         {
            n = Bearing::Attack(board, btwn_squares[i],
                                board.Side(), squares);
            for (unsigned j = 0; j < n; j++)
            {
               Piece p = board[squares[j]];
               if (TypeOfPiece(p) != King &&
                   !is_pinned(board, squares[j], btwn_squares[i]))
               {
                  Square dest = btwn_squares[i];
                  if (TypeOfPiece(p) == Pawn &&
                      Rank(dest,board.Side()) == 8)
                  {
                     // interposition is a promotion
                     moves[num_moves++] = CreateMove(board,
                                                     squares[j], dest, Queen);
                     moves[num_moves++] = CreateMove(board,
                                                     squares[j], dest, Rook);
                     moves[num_moves++] = CreateMove(board,
                                                     squares[j], dest, Knight);
                     moves[num_moves++] = CreateMove(board,
                                                     squares[j], dest, Bishop);
                  }
                  else
                  {
                     moves[num_moves++] = CreateMove(board, squares[j], dest, InvalidPiece);
                  }
               }
            }
         }
      }
   }
   if (captures_only)
      return num_moves;
   // generate evasive moves
   n = Bearing::BearSq(board, kp, squares);
//   Square atck_squares[8];
//   unsigned num_attackers = Bearing::Attack(board, kp,
//					board.OppositeSide(), atck_squares);
   for (i = 0; i < n; i++)
   {
      if (!board.any_attacks(squares[i], board.OppositeSide()))
         if (num_attacks > 1 || squares[i] != source)
         {
            // We need to do some extra checking, since the board
            // info on attacks reflects the state before the move,
            // and we need to be sure that the destination square
            // is not attacked after the move.
	   int illegal = 0;
            Bitmap it(king_attacks);
            Square atck_sq;
            while ((atck_sq = it.first_one()) != InvalidSquare && !illegal)
            {
               Piece attacker = board[(Square)(atck_sq)];
               if (sliding(attacker))
               {
                  int dir = Bearing::Directions[atck_sq][squares[i]];
                  // check for movement in the direction of the
                  // attacker:
                  if (dir != 0 && (kp + dir == squares[i]))
		    illegal = 1;
               }
               it.clear(atck_sq);
            }
            if (!illegal)
            {
               moves[num_moves++] = CreateMove(board, kp, squares[i], InvalidPiece);
            }
         }
   }
   return num_moves;
}
